// StripHistoryXMLRPC.cc
//
// Glue code between StripTool
// and the ChannelArchiver's network data server
//
// kasemir@lanl.gov

// StripMisc & string.h define basename.
// This define excludes the one from string.h, at least under RH9.
#define basename basename

// EPICS Base
#include <alarm.h>
// ChannelArchiver
#include <MsgLogger.h>
#include <stdString.h>
#include <epicsTimeHelper.h>
#include <ArchiveDataClient.h>
// StripTool
extern "C"
{
#include "StripHistory.h"
#include "StripDataSource.h"
}
    
// Define this to enable debug messages to the console
#undef DEBUG_STRIP_HISTORY

// Looks like StripTool will invoke StripHistory_fetch
// a) for a new channel
// b) when Panning to a new area
// c) when zooming in
//
// This code will always request new binned data.
// When zooming in, we request new binned data
// because otherwise one would see the effect of the
// binning in the detail.
// When zooming out, StripHistory_fetch isn't called,
// and that's fine.

// Number of bins.
// Assume a screen 1024...1280 pixel wide.
// StripTool will typically run as one tool
// amongst others, not using the full screen.
// Assume ~700 pixel for th width of the StripTool
// graph window. Minus legend and y axis, that might
// leave around 600 pixels for the data in the plot.
// ==> request 600 bins, resulting in 0 .. 600*4
//     data points.
#define NUM_BINS 600
#define BUF_COUNT (4*NUM_BINS)

// Instance data for the archive service.
class StripHistoryInfo
{
public:
    Strip         strip;
    stdString     url;
    int           key;
    ArchiveDataClient *client;
};

static bool epics_time_helper_init = false;

StripHistory StripHistory_init(Strip strip)
{
#ifdef DEBUG_STRIP_HISTORY
    LOG_MSG("StripHistory_init\n");
#endif    
    if (!epics_time_helper_init)
    {
        initEpicsTimeHelper();
        epics_time_helper_init = true;
    }
    const char *ARCHIVE_SERVER = getenv("ARCHIVE_SERVER");
    if (!ARCHIVE_SERVER)
    {
        fprintf(stderr, "Declare environment variable ARCHIVE_SERVER\n");
        fprintf(stderr, "to access archived data:\n");
        fprintf(stderr, "ARCHIVE_SERVER=URL:key\n");
        return 0;
    }
    const char *c = strrchr(ARCHIVE_SERVER, ':');
    if (!c)
    {
        fprintf(stderr, "ARCHIVE_SERVER does not match the format URL:key\n");
        return 0;
    }
    StripHistoryInfo *shi = new StripHistoryInfo;;
    if (!shi)
        return 0;
    shi->strip = strip;
    shi->url.assign(ARCHIVE_SERVER, c-ARCHIVE_SERVER);
    shi->key = atoi(c+1);
#ifdef DEBUG_STRIP_HISTORY
    LOG_MSG("ARCHIVE_SERVER='%s'\n", ARCHIVE_SERVER);
    LOG_MSG("URL: '%s'\n", shi->url.c_str());
    LOG_MSG("Key: %d\n", shi->key);
#endif
    shi->client = new ArchiveDataClient(shi->url.c_str());
    int version;
    stdString description;
    stdVector<stdString> how_strings, stat_strings;
    stdVector<ArchiveDataClient::SeverityInfo> sevr_infos;
    if (shi->client->getInfo(version, description,
                             how_strings, stat_strings, sevr_infos))
    {
        fprintf(stderr, "Connected to ArchiveDataServer V%d\n", version);
    }
    else
    {
        fprintf(stderr, "Cannot connect to '%s'\n", shi->url.c_str());
        delete shi->client;
        shi->client = 0;
    }
    return (StripHistory)shi;
}

void StripHistory_cancel(StripHistory sh, StripHistoryResult *result)
{
#ifdef DEBUG_STRIP_HISTORY
    LOG_MSG("StripHistory_cancel\n");
#endif
}

void StripHistory_delete(StripHistory sh)
{
#ifdef DEBUG_STRIP_HISTORY
    LOG_MSG("StripHistory_delete\n");
#endif
    StripHistoryInfo *shi = (StripHistoryInfo *)sh;
    delete shi->client;
    delete shi;
}

void  StripHistoryResult_release(StripHistory sh, StripHistoryResult *result)
{
#ifdef DEBUG_STRIP_HISTORY
    printf("StripHistory_release\n");
    LOG_MSG("result->data: 0x%08X\n", result->data);
#endif
    free(result->status);
    free(result->times);
    free(result->data);
    result->status = 0;
    result->times = 0;
    result->data = 0;
    result->n_points = 0;
    result->fetch_stat = FETCH_NODATA;
}

typedef struct
{
    size_t              buf_count;
    StripHistoryResult *result;
} CallbackInfo;

static bool value_callback(void *arg,
                           const char *name,
                           size_t ch_idx,
                           size_t idx,
                           const CtrlInfo &info,
                           DbrType type, DbrCount count,
                           const RawValue::Data *value)
{
    CallbackInfo *cb_info = (CallbackInfo *) arg;
    size_t       i = cb_info->result->n_points;
    bool         go_on = true;
    if (i >= cb_info->buf_count)
    {
        LOG_MSG("StripHistoryXMLRPC(%s): reached %d values\n", name, i);
        return false;
    }
    cb_info->result->times[i] = RawValue::getTime(value);
    if (RawValue::getDouble(type, count, value, cb_info->result->data[i]))
    {
        switch (RawValue::getSevr(value))
        {
            case INVALID_ALARM:
            case ARCH_DISCONNECT:
            case ARCH_STOPPED:
                cb_info->result->status[i] = 0;
                break;
            default:
                cb_info->result->status[i] = DATASTAT_PLOTABLE;
        }
    }
    else
    {
        cb_info->result->status[i] = 0;
        go_on = false;
    }
    cb_info->result->n_points = i+1;
    return go_on;
}

FetchStatus StripHistory_fetch(StripHistory sh, char *name,
                               struct timeval *t0, struct timeval *t1,
                               StripHistoryResult *result,
                               StripHistoryCallback callback,
                               void *callback_arg)
{
    StripHistoryInfo *shi = (StripHistoryInfo *)sh;
    stdVector<stdString> names;
    names.push_back(stdString(name));
    epicsTime start = *t0;
    epicsTime end = *t1;
#ifdef DEBUG_STRIP_HISTORY
    stdString txt;
    LOG_MSG("StripHistory_fetch '%s'\n", name);
    LOG_MSG("Start: %s\n", epicsTimeTxt(start, txt));
    LOG_MSG("End:   %s\n", epicsTimeTxt(end, txt));
#endif
    if (result->data)
        StripHistoryResult_release(sh, result);
    if (!(shi && shi->client))
        return FETCH_NODATA;
    result->data = (double *)         malloc(BUF_COUNT*sizeof(double));
    result->times = (struct timeval *)malloc(BUF_COUNT*sizeof(struct timeval));
    result->status = (short *)        malloc(BUF_COUNT*sizeof(short));
    CallbackInfo info;
    info.buf_count = BUF_COUNT;
    info.result = result;
    if (shi->client->getValues(shi->key, names, start, end,
                               NUM_BINS, HOW_PLOTBIN,
                               value_callback, (void *)&info))
    {
        result->t0 = start;
        result->t1 = end;
        size_t i = result->n_points;
        if (i > 0)
        {
            if (i < info.buf_count)
            {   // Keep StripTool from continuing beyond last data point
                result->times[i]  = result->times[i-1];
                result->status[i] = 0; // "not plottable"
                result->n_points  = i+1;
            }
            result->fetch_stat = FETCH_DONE;
        }
        else
            result->fetch_stat = FETCH_NODATA;
    }
    else
    {
        result->n_points = 0;
        result->fetch_stat = FETCH_NODATA;
    }
#ifdef DEBUG_STRIP_HISTORY
    LOG_MSG("result->data: 0x%08X, %d points\n",
            result->data, result->n_points);
    LOG_MSG("-------------------------------------\n");
#endif
    return result->fetch_stat;
}
 
